#include <yaco/config.hpp>

using yaco::config::Value;

Value::Value(const std::string& value)
        : value_(value)
{
}

Value::Value(std::int32_t value)
        : value_(value)
{
}

Value::Value(bool value)
        : value_(value)
{
}

Value::Value(const std::map<std::string, Value>& value)
        : value_(value)
{
}

Value::Value(const std::vector<Value>& value)
        : value_(value)
{
}

const std::string& Value::asString() const
{
    return std::any_cast<const std::string&>(value_);
}

const std::map<std::string, Value>& Value::asMap() const
{
    return std::any_cast<const std::map<std::string, Value>&>(value_);
}

const std::vector<Value>& Value::asArray() const
{
    return std::any_cast<const std::vector<Value>&>(value_);
}

std::int32_t Value::asInt32() const
{
    const std::string& str = std::any_cast<const std::string&>(value_);
    try {
        return std::stoi(str);
    } catch (const std::invalid_argument& e) {
        throw std::runtime_error("Value is not an integer");
    }
}

float Value::asFloat() const
{
    const std::string& str = std::any_cast<const std::string&>(value_);
    try {
        return std::stof(str);
    } catch (const std::invalid_argument& e) {
        throw std::runtime_error("Value is not a float");
    }
}

bool Value::asBool() const
{
    const std::string& str = std::any_cast<const std::string&>(value_);
    if (str == "true") {
        return true;
    } else if (str == "false") {
        return false;
    } else {
        throw std::runtime_error("Value is not a boolean");
    }
}

bool Value::isMap() const noexcept
{
    return std::any_cast<std::map<std::string, Value>>(&value_) != nullptr;
}

bool Value::isArray() const noexcept
{
    return std::any_cast<std::vector<Value>>(&value_) != nullptr;
}

bool Value::isScalar() const noexcept
{
    return std::any_cast<std::string>(&value_) != nullptr;
}

void Value::set(const std::string& value)
{
    value_ = value;
}

void Value::set(const std::map<std::string, Value>& value)
{
    value_ = value;
}

void Value::set(const std::vector<Value>& value)
{
    value_ = value;
}

void Value::set(std::int32_t value)
{
    value_ = std::to_string(value);
}

void Value::set(bool value)
{
    value_ = value ? "true" : "false";
}

Value& Value::operator[](const std::string& key) noexcept
{
    return (*std::any_cast<std::map<std::string, Value>>(&value_))[key];
}

Value& Value::at(const std::string& key)
{
    auto it = find(key);
    if (it == mapEnd()) {
        throw std::out_of_range("Key '" + key + "' not found");
    }
    return it->second;
}

const Value& Value::at(const std::string& key) const
{
    auto it = find(key);
    if (it == mapEnd()) {
        throw std::out_of_range("Key '" + key + "' not found");
    }
    return it->second;
}

Value::MapIterator Value::mapBegin()
{
    return std::any_cast<std::map<std::string, Value>&>(value_).begin();
}

Value::MapIterator Value::mapEnd()
{
    return std::any_cast<std::map<std::string, Value>&>(value_).end();
}

Value::ConstMapIterator Value::mapBegin() const
{
    return std::any_cast<const std::map<std::string, Value>&>(value_).begin();
}

Value::ConstMapIterator Value::mapEnd() const
{
    return std::any_cast<const std::map<std::string, Value>&>(value_).end();
}

Value::MapIterator Value::find(const std::string& key)
{
    return std::any_cast<std::map<std::string, Value>&>(value_).find(key);
}

Value::ConstMapIterator Value::find(const std::string& key) const
{
    return std::any_cast<const std::map<std::string, Value>&>(value_).find(key);
}

bool Value::has(const std::string& key) const
{
    return find(key) != mapEnd();
}

/** Array Methods **/

Value& Value::operator[](std::size_t i) noexcept
{
    return (*std::any_cast<std::vector<Value>>(&value_))[i];
}

const Value& Value::operator[](std::size_t i) const noexcept
{
    return (*std::any_cast<const std::vector<Value>>(&value_))[i];
}

Value& Value::at(std::size_t i)
{
    return std::any_cast<std::vector<Value>&>(value_).at(i);
}

const Value& Value::at(std::size_t i) const
{
    return std::any_cast<const std::vector<Value>&>(value_).at(i);
}

Value::ArrayIterator Value::arrayBegin()
{
    return std::any_cast<std::vector<Value>&>(value_).begin();
}

Value::ArrayIterator Value::arrayEnd()
{
    return std::any_cast<std::vector<Value>&>(value_).end();
}

Value::ConstArrayIterator Value::arrayBegin() const
{
    return std::any_cast<const std::vector<Value>&>(value_).begin();
}

Value::ConstArrayIterator Value::arrayEnd() const
{
    return std::any_cast<const std::vector<Value>&>(value_).end();
}
