#include "fdb_writer.hpp"

#include <eccodes.h>
#include <eckit/runtime/Main.h>
#include <fdb5/api/FDB.h>
#include <fdb5/config/Config.h>
#include <yaco/chrono.hpp>
#include <yaco/collection.hpp>
#include <yaco/config.hpp>
#include <yaco/grid.hpp>

using yaco::FDBWriter;
using yaco::Grid;
using yaco::chrono::DateTime;

#ifndef YACO_GRIB_MISSING_VALUE
#define YACO_GRIB_MISSING_VALUE -9.e33
#endif

namespace config = yaco::config;

std::unique_ptr<yaco::DataHandler> yaco::DataHandler::createFDBWriter(const config::Value& doc)
{
    return std::make_unique<FDBWriter>(doc);
}

static std::vector<std::uint8_t> bytesFromHexString(const std::string& hex)
{
    std::vector<std::uint8_t> bytes;
    for (std::size_t i = 0; i < hex.size(); i += 2) {
        bytes.push_back(static_cast<std::uint8_t>(std::stoi(hex.substr(i, 2), nullptr, 16)));
    }
    return bytes;
}

static std::string hexStringFromBytes(const std::vector<std::uint8_t>& bytes)
{
    std::string str(bytes.size() * 2, ' ');
    for (std::size_t i = 0; i < bytes.size(); ++i) {
        sprintf(&str[2 * i], "%02x", bytes[i]);
    }
    return str;
}

class FDBWriter::GRIBMessage {
  private:
    codes_handle* handle_ = nullptr;

  public:
    GRIBMessage()
    {
        handle_ = codes_grib_handle_new_from_samples(0, "GRIB2");
        if (handle_ == nullptr) {
            throw std::runtime_error("Failed to create eccodes handle");
        }
    }
    ~GRIBMessage()
    {
        codes_handle_delete(handle_);
    }

    void set(const DateTime& dt)
    {
        GRIB_CHECK(codes_set_long(handle_, "dataDate", dt.year() * 10000 + dt.month() * 100 + dt.day()), 0);
        GRIB_CHECK(codes_set_long(handle_, "hour", dt.hour()), 0);
        GRIB_CHECK(codes_set_long(handle_, "minute", dt.minute()), 0);
        GRIB_CHECK(codes_set_long(handle_, "second", dt.second()), 0);
    }

    void set(const Field& field)
    {
        CODES_CHECK(codes_set_double_array(handle_, "values", field.data(), field.size()), 0);
    }

    void set(const std::string& key, const std::string& value)
    {
        std::size_t len = value.size();
        int res         = codes_set_string(handle_, key.c_str(), value.c_str(), &len);
        if (res != 0) {
            throw std::runtime_error("Failed to set key '" + key + "' to string '" + value + "': " + codes_get_error_message(res));
        }
    }

    void set(const std::string& key, const std::vector<std::uint8_t>& value)
    {
        std::size_t len = value.size();
        int res         = codes_set_bytes(handle_, key.c_str(), (const unsigned char*)&value[0], &len);
        if (res != 0) {
            throw std::runtime_error("Failed to set key '" + key + "' to string '" + hexStringFromBytes(value) + "': " + codes_get_error_message(res));
        }
    }

    void set(const std::string& key, double value)
    {
        int res = codes_set_double(handle_, key.c_str(), value);
        if (res != 0) {
            throw std::runtime_error("Failed to set key '" + key + "' to double '" + std::to_string(value) + "': " + codes_get_error_message(res));
        }
    }

    void set(const std::string& key, long value)
    {
        int res = codes_set_long(handle_, key.c_str(), value);
        if (res != 0) {
            throw std::runtime_error("Failed to set key '" + key + "' to long '" + std::to_string(value) + "': " + codes_get_error_message(res));
        }
    }

    void archive(fdb5::FDB& fdb, const fdb5::Key& key)
    {
        const void* data;
        std::size_t data_length;
        int res = codes_get_message(handle_, &data, &data_length);
        if (res != 0) {
            throw std::runtime_error("Failed to get raw message: " + std::string(codes_get_error_message(res)));
        }
        fdb.archive(key, data, data_length);
    }
};

class FDBWriter::ParamCopier {

    bool handleSpecialKeys(GRIBMessage& msg, const std::string& key, const config::Value& value)
    {
        if (key == "uuidOfVGrid") {
            msg.set(key, bytesFromHexString(value.asString()));
            return true;
        }
        return false;
    }

    void setConfigValue(GRIBMessage& msg, const std::string& key, const config::Value& value)
    {
        if (handleSpecialKeys(msg, key, value)) {
            return;
        }

        bool hasPoint    = false;
        bool isNumerical = (std::find_if(value.asString().begin(), value.asString().end(),
                                         [&hasPoint](char c) {
                                             if (!std::isdigit(c)) {
                                                 if (!hasPoint && c == '.') {
                                                     hasPoint = true;
                                                     return false;
                                                 }
                                                 return true;
                                             }
                                             return false;
                                         })
                            == value.asString().end());
        if (isNumerical) {
            if (hasPoint) {
                msg.set(key, (double)value.asFloat());
            } else {
                msg.set(key, (long)value.asInt32());
            }
        } else {
            msg.set(key, value.asString());
        }
    }

    void setConfigValue(fdb5::Key& fdbKey, const std::string& key, const config::Value& value)
    {
        fdbKey.set(key, value.asString());
    }

    template<class Dict>
    void copyFromConfigToDict(const config::Value& gribConfig, Dict& msg)
    {
        for (auto iter = gribConfig.mapBegin(); iter != gribConfig.mapEnd(); ++iter) {
            if (iter->second.isScalar()) {
                setConfigValue(msg, iter->first, iter->second);
            } else if (iter->second.isMap()) {
                const auto use = iter->second.at("use").asString();
                ParamMap params;
                for (auto paramIter = iter->second.mapBegin(); paramIter != iter->second.mapEnd(); ++paramIter) {
                    if (paramIter->first != "use") {
                        params[paramIter->first] = paramIter->second.asString();
                    }
                }
                setConfigValue(msg, iter->first, handlers_.at(use)(params));
            }
        }
    }

  public:
    using ParamMap     = std::map<std::string, std::string>;
    using ParamHandler = std::function<std::string(const ParamMap&)>;
    ParamCopier()      = default;
    ~ParamCopier()     = default;

    void bind(const std::string& param, ParamHandler handler)
    {
        handlers_[param] = std::move(handler);
    }

    void copyFromConfigToGRIBMessage(const config::Value& gribConfig, GRIBMessage& msg)
    {
        copyFromConfigToDict(gribConfig, msg);
    }

    void copyFromConfigToFDBKey(const config::Value& gribConfig, fdb5::Key& fdbKey)
    {
        copyFromConfigToDict(gribConfig, fdbKey);
    }

  private:
    std::map<std::string, ParamHandler> handlers_;
};

class FDBWriter::Impl {
  public:
    Impl(const config::Value& config) : config_(config)
    {
        if (!fdb_lib_initialized_) {
            std::string arg = "fdb";
            std::array<char*, 2> argv({ arg.data(), nullptr });
            eckit::Main::initialise(argv.size() - 1, argv.data());
            fdb_lib_initialized_ = true;
        }

        const auto fdb_config = config_.find("fdb_config");
        if (fdb_config != config_.mapEnd()) {
            fdb_ = std::make_unique<fdb5::FDB>(fdb5::Config::make(fdb_config->second.asString()));
        } else {
            fdb_ = std::make_unique<fdb5::FDB>();
        }
    }

    void handle(Collection& collection, const DateTime& timestep);

  private:
    static bool fdb_lib_initialized_;
    config::Value config_;
    std::unique_ptr<fdb5::FDB> fdb_;
};

bool FDBWriter::Impl::fdb_lib_initialized_ = false;

FDBWriter::FDBWriter(const config::Value& config)
        : impl_(std::make_unique<Impl>(config))
{
}

FDBWriter::~FDBWriter() = default;

void FDBWriter::Impl::handle(Collection& collection, const DateTime& datetime)
{
    const auto& collectionNode = config_.at("collections").at(collection.name());

    for (std::size_t i = 0; i < collection.size(); ++i) {
        GRIBMessage msg;
        msg.set(datetime);

        // Reset NANs to missing value
        bool haveMissingValues = false;
        for (std::size_t valueIdx = 0; valueIdx < collection.at(i).size(); ++valueIdx) {
            double& value = collection.at(i).at(valueIdx);
            if (std::isnan(value)) {
                value             = YACO_GRIB_MISSING_VALUE;
                haveMissingValues = true;
            }
        }
        if (haveMissingValues) {
            msg.set("missingValue", YACO_GRIB_MISSING_VALUE);
            msg.set("bitmapPresent", (long)1);
        }

        msg.set(collection.at(i));

        ParamCopier copier;
        copier.bind("datetime", [&datetime](const ParamCopier::ParamMap& params) { return datetime.toString(params.at("datetime").c_str()); });
        copier.bind("grid", [&grid = collection.grid()](const ParamCopier::ParamMap&) { return grid.name(); });
        copier.bind("collection_index", [&i](const ParamCopier::ParamMap& params) { return std::to_string(static_cast<long>(i) + (params.find("add") == params.end() ? 0 : std::stol(params.at("add")))); });
        copier.copyFromConfigToGRIBMessage(config_.at("grib"), msg);
        copier.copyFromConfigToGRIBMessage(collectionNode.at("grib"), msg);

        fdb5::Key key{};
        copier.copyFromConfigToFDBKey(config_.at("fdb_key"), key);
        copier.copyFromConfigToFDBKey(collectionNode.at("fdb_key"), key);
        msg.archive(*fdb_, key);
    }
}

void FDBWriter::handle(Collection& collection, const DateTime& datetime)
{
    impl_->handle(collection, datetime);
}
